/*
 * gc global variable provides access to GUI Composer infrastructure components and project information.
 * For more information, please see the Working with Javascript guide in the online help.
 */
var gc = gc || {};
gc.services = gc.services || {};

var rtls = rtls || {};
var scanResults = [];
var requestHandler = null;


const ws_send = (req) => {
    console.log('>> ' + JSON.stringify(req))
    gc.ws.send(JSON.stringify(req))
};

const log = (level, msg) => {
    console.log(msg);
    if (templateObj) templateObj.$.logview.log(level, msg)
}

const controlApi = {
    list_devices: () => ws_send({ control: { req: 'LIST_DEVICES' } })
};

const syncreq = (identifier, command, payload) => ({ identifier, message: { type: "SyncReq", subsystem: "RTLS", command, payload: payload || {} } })

const Device = (devInfo) => {
    const state = {};

    return {
        ...devInfo,
        isMaster: devInfo.caps.indexOf('RTLS_MASTER') > -1,
        identify: () => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_IDENTIFY')),
        reset: () => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_RESET_DEVICE')),
        scan: () => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_SCAN')),
        connect: (addrType, peerAddr) => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_CONNECT', { addrType, peerAddr })),
        terminate: () => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_TERMINATE_LINK')),
        setConnInfo: (accessAddress, connInterval, hopValue, mSCA, currChan, chanMap) => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_CONN_PARAMS', { accessAddress, connInterval, hopValue, mSCA, currChan, chanMap })),
        //
        tofStart: (enable) => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_TOF_ENABLE', { enable })),
        tofSetParams: (tofRole, numSamples, numFreq, autoTofRssiThresh, resultMode, runMode, frequencies) => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_TOF_SET_PARAMS', { tofRole, numSamples, numFreq, autoTofRssiThresh, resultMode, runMode, frequencies })),
        tofSetSeed: (seed) => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_TOF_SET_SEC_SEED', { seed })),
        tofGetSeed: () => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_TOF_GET_SEC_SEED')),
        //
        aoaStart: (enable) => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_AOA_ENABLE', { enable })),
        aoaSetParams: (aoaRole, aoaResultMode, cteScanOvs, cteOffset, cteTime) => ws_send(syncreq(devInfo.identifier, 'RTLS_CMD_AOA_SET_PARAMS', { aoaRole, aoaResultMode, cteScanOvs, cteOffset, cteTime })),
        //
        state,
    }
}

const Request = (msg) => {
    const device = rtls.devices && rtls.devices.find(e => e.identifier === msg.identifier);
    return {
        device: device,
        ...msg.message,
    }
}


const handle_control = (control) => {
    switch (control.req) {
        case 'LIST_DEVICES':
            rtls.devices = control.devices.map(e => Device(e));

            rtls.master = rtls.devices.find(e => e.isMaster);
            rtls.passives = rtls.devices.filter(e => !e.isMaster);

            // Error check and mode check
            // 1. Need at least 1 master
            if (!rtls.master) log('error', 'Need a master device');
            
            // 2. All must be either AoA or ToF
            const allTof = rtls.devices.every(e => (e.caps.indexOf('TOF_MASTER') >= 0) || (e.caps.indexOf('TOF_PASSIVE') >= 0));
            const allNotTof = rtls.devices.every(e => !((e.caps.indexOf('TOF_MASTER') >= 0) || (e.caps.indexOf('TOF_PASSIVE') >= 0)));
            if (!(allTof || allNotTof)) log('error', 'All nodes must be either AoA or ToF');
            
            // 3. Need at least 1 passive for AoA
            if (allNotTof && rtls.passives.length === 0) log('error', 'Need at least 1 passive for AoA');
           
            // 4. Set up the command handler
            if (allTof) requestHandler = handle_request_tof;
            else requestHandler = handle_request_aoa;

            // Initialize graph
            if (templateObj) {
                const deviceLabels = control.devices.map(e => `${e.port} / ${e.identifier} / ${e.caps.join('+')}`).join('|');
                templateObj.$.lst_connected_devices.labels=deviceLabels;

                const graphNodes = allTof ? [rtls.master].concat(rtls.passives.slice(0,3)) : rtls.passives.slice(0,4);
                const graph = templateObj.$.scalargraph;
                graph.titleLabel = allTof ? 'Time of Flight' : 'Angle of Arrival';
                graph.fixedScale = true;
                graph.numChannels = graphNodes.length;
                graphNodes.forEach((e, i) => {
                    e.graphId = i;
                    graph['series_'+i+'_label'] = e.caps.join('+')
                    graph['series_'+i+'_units'] = allTof ? 'meters' : 'degrees';
                    graph['series_'+i+'_min'] = allTof ? 0 : -150;
                    graph['series_'+i+'_max'] = allTof ? 30 : 150;
                })
            }

            // Initialize devices
            scanResults = [];

            rtls.master.terminate();
            rtls.passives.forEach(it => it.reset());
            setTimeout(() => {
                rtls.master.reset();
                setTimeout(() => {
                    rtls.master.scan();
                }, 500);
            }, 500);
            break;
        default:
            log('error', 'Unknown response: ' + control.req);
            break;
    }
}

let peerAddr;
let peerAddrType;
let seed;



const handle_request_aoa = (req) => {
    //
    // Connection establishment to remote device
    //

    // Save latest scan response
    if (req.command === 'RTLS_CMD_SCAN' && req.type === 'AsyncReq') {
        peerAddr = req.payload.addr;
        peerAddrType = req.payload.addrType;
        gui_scan_result(req);
    }

    // Connect to last found when scan is finished
    if (req.command === 'RTLS_CMD_SCAN_STOP') {
        if (peerAddr) {
            rtls.master.connect(peerAddrType, peerAddr);
        } else {
            rtls.master.scan();
        }
    }

    // When connected, let passives know about connection params
    if (req.command === 'RTLS_CMD_CONN_PARAMS' && req.type === 'AsyncReq') {
        const { accessAddress, connInterval, hopValue, mSCA, currChan, chanMap } = req.payload;
        rtls.passives.forEach(e => {
            e.setConnInfo(accessAddress, connInterval, hopValue, mSCA, currChan+hopValue*1, chanMap);
        });
    }

    // When connection parameters are accepted by passive, note this down
    if (req.command === 'RTLS_CMD_CONN_PARAMS' && req.type === 'SyncRsp')
    {
        req.device.state.connSet = true;
    }

    //
    // Initialize AoA
    //

    // When master sends CONNECT later, pairing is complete and we can initialize AoA parameters
    if (req.command === 'RTLS_CMD_CONNECT' && req.type === 'AsyncReq' && req.payload.status === 'RTLS_SUCCESS' && req.device === rtls.master) {
        rtls.master.aoaSetParams('AOA_MASTER', 'AOA_MODE_ANGLE', 4, 4, 20);
        rtls.passives.forEach(it => {
           it.aoaSetParams('AOA_PASSIVE', 'AOA_MODE_ANGLE', 4, 4, 20);
        })
    }

    // After aoa parameters are set, start aoa
    if (req.command === 'RTLS_CMD_AOA_SET_PARAMS') {
        req.device.aoaStart(1);
    }

    //
    // Receive Data
    //

    if (req.command === 'RTLS_CMD_AOA_RESULT_ANGLE')
    {
        if (templateObj) {
            const graph = templateObj.$.scalargraph;
            const id = req.device.graphId;
            if (id !== undefined) {
                graph['value_'+id.toString()] = req.payload.angle;
            }
        }
    }
}


const handle_request_tof = (req) => {
    //
    // Connection establishment to remote device
    //

    // Save latest scan response
    if (req.command === 'RTLS_CMD_SCAN' && req.type === 'AsyncReq') {
        peerAddr = req.payload.addr;
        peerAddrType = req.payload.addrType;
        gui_scan_result(req);
    }

    // Connect to last found when scan is finished
    if (req.command === 'RTLS_CMD_SCAN_STOP') {
        if (peerAddr) {
            rtls.master.connect(peerAddrType, peerAddr);
        } else {
            rtls.master.scan();
        }
    }

    // When connected, let passives know about connection params
    if (req.command === 'RTLS_CMD_CONN_PARAMS' && req.type === 'AsyncReq') {
        const { accessAddress, connInterval, hopValue, mSCA, currChan, chanMap } = req.payload;
        rtls.passives.forEach(e => {
            e.setConnInfo(accessAddress, connInterval, hopValue, mSCA, currChan+hopValue*1, chanMap);
        });
    }

    // When connection parameters are accepted by passive, note this down
    if (req.command === 'RTLS_CMD_CONN_PARAMS' && req.type === 'SyncRsp')
    {
        req.device.state.connSet = true;
    }

    //
    // Initialize ToF
    //

    // When master sends CONNECT later, pairing is complete and we can initialize ToF parameters
    if (req.command === 'RTLS_CMD_CONNECT' && req.type === 'AsyncReq' && req.payload.status === 'RTLS_SUCCESS' && req.device === rtls.master) {
        rtls.master.tofSetParams('TOF_MASTER', 64, 1, -55, 'TOF_MODE_DIST', 'TOF_MODE_CONT', [2402]);
        rtls.passives.forEach(it => {
           it.tofSetParams('TOF_PASSIVE', 64, 1, -55, 'TOF_MODE_DIST', 'TOF_MODE_CONT', [2402]);
        })
    }

    // After tof_init, get the random seed from the master
    if (req.command === 'RTLS_CMD_TOF_SET_PARAMS') {
        req.device.state.tofSet = true;
        
        if (req.device === rtls.master){
            rtls.master.tofGetSeed();
        }
    }
    
    // When we get seed, give it to the passives
    if (req.command === 'RTLS_CMD_TOF_GET_SEC_SEED') {
        seed = req.payload.seed;
        rtls.passives.forEach(e => {
            if (e.state.tofSet) {
                e.tofSetSeed(seed);
            } else {
                console.error("Passive is not initialized");
            }
        })
    }

    // When all seeds are set, start ToF
    if (req.command === 'RTLS_CMD_TOF_SET_SEC_SEED') {
        req.device.state.seedSet = true;
        if (rtls.passives.every(e => e.state.seedSet)) {
            if (rtls.passives.every(e => e.state.connSet)) {
                rtls.passives.forEach(e => {
                    e.tofStart(1);
                })
                rtls.master.tofStart(1);
            }
    
        }
    }

    //
    // Receive Data
    //

    if (req.command === 'RTLS_CMD_TOF_RESULT_DIST')
    {
        if (templateObj) {
            const graph = templateObj.$.scalargraph;
            const id = req.device.graphId;
            if (id !== undefined) {
                graph['value_'+id.toString()] = req.payload.distance;
            }
        }
    }
}


const ws_receive = (msg) => {
    // console.log(msg.data);
    if (templateObj) templateObj.$.logview.log('info', msg.data);
    try {
        const parsed = JSON.parse(msg.data)

        if (parsed.control) {
            handle_control(parsed.control);
        } else if (rtls.devices) {
            const req = Request(parsed);
            if (req.command === 'RTLS_CMD_ERROR' && req.payload.status === 'RTLS_ILLEGAL_CMD') log('error', 'Illegal command received. Probably tried to do ToF on an AoA node or vice versa');
            // console.log(`${req.command} from ${req.device.name}`);

            if (!requestHandler) {
                log('error', 'No request handler, have not detected AoA or ToF nodes connected');
                return;
            }
            
            requestHandler(req);
        }
    } catch (err) {
        log('error', err);
    }
}


const wss_connect = () => {
    try {
        var ws_addr = templateObj.$.input_ws_addr.value
        gc.ws = new WebSocket(ws_addr);
        gc.ws.onmessage = ws_receive;
        gc.ws.onclose = () => {
            templateObj.$.logview.log('error', "Closed");
            templateObj.$.btn_connect.disabled = false;
        }
        gc.ws.onopen = () => {
            controlApi.list_devices();
        }
        templateObj.$.btn_connect.disabled = true;
    } catch (err) {
        templateObj.$.logview.log('error', "Could not connect")
        alert("Could not connect!");
        console.log(err);
    }
}

gc.tofs = gc.tofs || [];


const gui_scan_result = (req) => {
    if (!templateObj) return;
    const list = templateObj.$.lst_remote_devices;

    scanResults.push({...req.payload});
    list.labels = scanResults.map(e => `${e.addr}`).join('|');
}



/*
*  Boilerplate code for creating computed data bindings
*/
document.addEventListener('gc-databind-ready', function() 
{
    /* 
	*   Add custom computed value databindings here, using the following method:
	*
    *   function gc.databind.registry.bind(targetBinding, modelBinding, [getter], [setter]);
	*
	*   param targetBinding - single binding string or expression, or array of binding strings for multi-way binding.
	*   param modelBinding - single binding string or expression, or array of binding strings for multi-way binding.
	*   param getter - (optional) - custom getter function for computing the targetBinding value(s) based on modelBinding value(s).
	*   param setter - (optional) - custom setter function for computing the modelBinding value(s) based on targetBinding value(s).
    */
	
	// For example, a simple computed values based on simple expression
	// gc.databind.registry.bind('widget.id.propertyName', "targetVariable == 1 ? 'binding is one' : 'binding is not one'");
	
	// Or a custom two-way binding with custome getter and setter functions.  (setter is optional)  (getter only indicates one-way binding)
	// gc.databind.registry.bind('widget.id.propertyName', "targetVariable", function(value) { return value*5/9 + 32; }, function(value) { (return value-32)*9/5; });
	
	// Event 1 to n bindings
	/* 
    gc.databind.registry.bind('widget.date.value', 
        // dependant bindings needed in order to compute the date, in name/value pairs.
        {
            weekday: 'widget.dayOfWeek.selectedText',
            day: 'widget.dayOfMonth.value',
            month: 'widget.month.selectedText',
            year: 'widget.year.value'
        }, 
        // getter for date computation
        function(values) 
        {
            // compute and return the string value to bind to the widget with id 'date'
            return values.weekday + ', ' + values.month + ' ' + values.day + ', ' + values.year;
        }
    ); 
	*/
});

/*
*  Boilerplate code for creating custom actions
*/
document.addEventListener('gc-nav-ready', function() 
{
    /* 
	*   Add custom actions for menu items using the following api:
	*
    *   function gc.nav.registryAction(id, runable, [isAvailable], [isVisible]);
	*
	*   param id - uniquely identifies the action, and should correspond to the action property of the menuaction widget.
	*   param runable - function that performs the custom action.
	*   param isAvailable - (optional) - function called when the menu action is about to appear.  Return false to disable the action, or true to enable it.
	*   param isVisible - (optional) - function called when the menu action is about to appear.  Return false to hide the action, or true to make it visible.
    */
	
	// For example,
	// gc.nav.registerAction('myCustomCloseAction', function() { window.close(); }, function() { return true; }, function() { return true; });
	
	// Alternatively, to programmatically disable a menu action at any time use:
	// gc.nav.disableAction('myCustomCloseAction);    then enable it again using:  gc.nav.enableAction('myCustomCloseAction'); 
});


/*
*  Boilerplate code for working with components in the application gist
*/


var initComplete = false;
var templateObj;

// Wait for DOMContentLoaded event before trying to access the application template
var init = function() {
    templateObj = document.querySelector('#template_obj');

    // Wait for the template to fire a dom-change event to indicate that it has been 'stamped'
    // before trying to access components in the application.
	templateObj.addEventListener('dom-change',function(){
	    if (initComplete) return;
	    this.async(function(){
    	    initComplete = true;
    	    console.log("Application template has been stamped.");
  	        // Now that the template has been stamped, you can use 'automatic node finding' $ syntax to access widgets.
  	        // e.g. to access a widget with an id of 'widget_id' you can use templateObj.$.widgetId
  	        
  	        templateObj.$.btn_connect.onclick = wss_connect;
            templateObj.$.btn_connect.disabled = false;
            templateObj.$.chk_autortls.checked = true;
            templateObj.$.chk_autortls.disabled = true;
            
            templateObj.$.btn_scan.disabled = true;
            templateObj.$.btn_conn_disconn.disabled = true;
            templateObj.$.btn_enable_rtls.disabled = true;
              
            templateObj.$.btn_reload.onclick = () => document.location.reload();
  	           
        },1);

	});
};

templateObj = document.querySelector('#template_obj');
if (templateObj) {
    init();
} else {
    document.addEventListener('DOMContentLoaded',init.bind(this))
}


